#!/usr/bin/env python
from gnuradio import gr
from gnuradio import blocks
from gnuradio import digital
from gnuradio import uhd

from optparse import OptionParser
from gnuradio import eng_notation
from gnuradio.eng_option import eng_option

class my_top_block(gr.top_block):
    def __init__(self, rx_callback, freq):
        gr.top_block.__init__(self)

        ##################################################
        # Variables
        ##################################################
        self.freq = freq
        self.samp_rate = 1e6
        self.gain = 20

        ##################################################
        # Blocks
        ##################################################
        self.uhd_usrp_sink = uhd.usrp_sink(
            device_addr="",
            stream_args=uhd.stream_args(
                cpu_format="fc32",
                channels=range(1),
            ),
        )
        self.uhd_usrp_sink.set_samp_rate(self.samp_rate)
        self.uhd_usrp_sink.set_center_freq(self.freq, 0)        
        self.uhd_usrp_sink.set_gain(self.gain, 0)

        self.uhd_usrp_source = uhd.usrp_source(
            device_addr="",
            stream_args=uhd.stream_args(
                cpu_format="fc32",
                channels=range(1),
            ),
        )
        self.uhd_usrp_source.set_samp_rate(self.samp_rate)
        self.uhd_usrp_source.set_center_freq(self.freq, 0)
        self.uhd_usrp_source.set_gain(self.gain, 0)

        self.txpath = transmit_path()
        self.rxpath = receive_path(rx_callback)


        ##################################################
        # Connections
        ##################################################

        self.connect(self.txpath, self.uhd_usrp_sink)
        self.connect(self.uhd_usrp_source, self.rxpath)

    def send_pkt(self, payload='', eof=False):
        return self.txpath.send_pkt(payload, eof)

class transmit_path(gr.hier_block2):
    def __init__(self):
        gr.hier_block2.__init__(self, "transmit_path",
            gr.io_signature(0,0,0),
            gr.io_signature(1, 1, gr.sizeof_gr_complex))

        ##################################################
        # Variables
        ##################################################       
        self.msgq_limit = 2

        ##################################################
        # Blocks
        ##################################################
        
        self.pkt_input = blocks.message_source(gr.sizeof_char, msgq_limit)

        # self.input_unpacked_to_packed = blocks.unpacked_to_packed_bb(1, gr.GR_MSB_FIRST)

        self.mod = digital.dbpsk_mod(
            samples_per_symbol=2,
            excess_bw=0.35,
            mod_code="gray",
            verbose=False,
            log=False)

        ##################################################
        # Connections
        ##################################################
        self.connect(self.pkt_input, self.mod, self)

    def send_pkt(self, payload='', eof=False):
        if eof:
            msg = gr.message(1)  # tell self._pkt_input we're not sending any more packets
        else:
            msg = gr.message_from_string(payload)
        
        self.pkt_input.msgq().insert_tail(msg)

class receive_path(gr.hier_block2):
    def __init__(self, rx_callback):
        gr.hier_block2.__init__(self, "receive_path",
            gr.io_signature(1, 1, gr.sizeof_gr_complex),
            gr.io_signature(0, 0, 0))

        ##################################################
        # Variables
        ##################################################
        self.callback = rx_callback

        ##################################################
        # Blocks
        ##################################################
        self.demod = digital.dbpsk_demod(
            samples_per_symbol=2,
            excess_bw=0.35,
            freq_bw=6.28/100.0,
            phase_bw=6.28/100.0,
            timing_bw=6.28/100.0,
            mod_code="gray",
            verbose=False,
            log=False
        )        
               
        self.frame_sync = frame_sync() # Your custom block!!!

        self.output_unpacked_to_packed = blocks.unpacked_to_packed_bb(1, gr.GR_MSB_FIRST)

        self.rcvd_pktq = gr.msg_queue()
        self.message_sink = blocks.message_sink(gr.sizeof_char, self.rcvd_pktq, True)

        self.queue_watcher_thread = _queue_watcher_thread(self.rcvd_pktq, self.callback)

        ##################################################
        # Connections
        ##################################################
        self.connect(self, self.self.demod, self.frame_sync, self.output_unpacked_to_packed, self.message_sink)

class _queue_watcher_thread(_threading.Thread):
    def __init__(self, rcvd_pktq, callback):
        _threading.Thread.__init__(self)
        self.setDaemon(1)
        self.rcvd_pktq = rcvd_pktq
        self.callback = callback
        self.keep_running = True
        self.start()

    def run(self):
        while self.keep_running:
            msg = self.rcvd_pktq.delete_head()
            payload = msg.to_string()
            if self.callback:
                self.callback(payload)

class tdd_mac(object):
    def __init__(self):
        self.tb = None

    def set_top_block(self, tb):
        self.tb = tb

    def rx_callback(self, payload):
        # Stuff goes here like:
        # print payload

    def main_loop(self): 
        while 1:           
            ##################################################
            # Your Mac logic goes here                       #
            ##################################################
            # self.tb.send_pkt("Hello Wrld")
            # time.sleep(5)
            # The receiver thread automatically handles receiving


def main():

    parser = OptionParser(option_class=eng_option, conflict_handler="resolve")
    parser.add_option("-f", "--freq", type="eng_float", default=None,
                          help="set center frequency to FREQ", metavar="FREQ")
    parser.add_options("-m", type="eng_float", default=1,
                        help = "set mode, tx or rx")

    (options, args) = parser.parse_args()

    freq = options.freq
    print "Freq = " + str(freq)

    if freq is None:
        raise ValueError("Center frequency not specified...")

    r = gr.enable_realtime_scheduling()
    if r != gr.RT_OK:
        print "Warning: failed to enable realtime scheduling"

    mac = tdd_mac()

    tb = my_top_block(mac.rx_callback, freq)

    mac.set_top_block(tb)

    tb.start() # Start executing the flow graph (runs in separate threads)

    mac.main_loop() # don't expect this to return

    tb.stop() # but if it does, tell flow graph to stop.
    tb.wait() # wait for it to finish

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass
